#!/usr/bin/perl

#                 d s m r e c o n c i l e G P F S . p l
###############################################################################
#
# This is a sample TSM for Space Management script for the policy driven 
# reconciliation method. 
#
# The policy driven reconciliation exists of three steps:
#
# 1. use the GPFS policy engine to generate a list of all migrated files 
#    in the file system.
# 2. prepare the generated file list for the usage in the TSM command
#    dsmreconcile.
# 3. Invoke the dsmreconcile command with the arguments -orphan -filelist
#    and perform the reconciliation.
#
# This script demonstrates how these steps can be implemented in perl to 
# automate the whole reconiliation process.
#
###############################################################################

use strict;

# set global variables ########################################################

my $debug             = 0;

my $nFiles            = 0;
# Can be device or mounting point - accepted by mmapplypolicy
my $fsName            = "";
# Must be the mounting point (path)
my $fsMntPt           = "";
my $serverName        = "";
my $useServerSpecRule = 0;
my $returnCode        = 0;

my $startTime     = time;
# default that must be overridden with a GPFS path for higher file counts
my $tmpDir        = "/tmp";
my $scanOut       = "scan.out";
my $ruleFile;
my $ruleOutput;
my $scanOutFile;
my $prepOutput;
my $reorgOutput;

# get command line arguments ##################################################

my $scriptName = $0;

# declare sub routines ########################################################

sub verifyArguments
{
   my $opt;

   if (@ARGV == 0) {
      die "Usage: $scriptName [-tmpdir=<dir>] [-server=<server>] <filesystem>\n";
   }
 
   while(@ARGV)
   {
      $opt = shift @ARGV;
      last if (! defined $opt);
 
      if( $opt =~ m/^-server=(.*)/) 
      {
         $serverName = uc($1);
         if( length($serverName) == 0 ) {
            die "ERROR : you need to specify a valid TSM Server name\n";
         }
         $useServerSpecRule = 1;
      }
      elsif( $opt =~ m/^-tmpdir=(.*)/)
      {
         $tmpDir = $1;
         if( length($tmpDir) == 0 ) {
            die "ERROR : you need to specify a valid path for tmp files\n";
         }
         if( ! -d $tmpDir) {
            mkdir($tmpDir) or die "ERROR: Could not create $tmpDir";
         }
      }
      elsif( $opt eq '-d' or $opt eq '-debug' )
      {
         $debug += 1;
      }
      elsif( $opt =~ m/^-/)
      {
         die "No such option: $opt\n";
      }
      else
      {
         $fsName = $opt;
         if( $fsName =~ '^/dev/.+' ) {
            print "$fsName was recognized as a device name\n" if ($debug > 0);
            trDevToMntPoint();
            print "  $fsName was translated to path $fsMntPt\n" if ($debug > 0);
         } else {
            $fsMntPt = $fsName;
            if( verifyMntPoint() == 1 ) {
               print "$fsMntPt was recognized as a mounting point\n" if ($debug > 0);
            } else {
               die "Mounting point $fsMntPt does not exist.\n";
            }
         }
         if( ! -d "$fsMntPt" ) {
            die "$fsMntPt directory does not exist.\n";
         }
      }
   }

   if( length($fsName) == 0 ) {
      die "ERROR : File system must be specified.\n";
   }
  
   my $gpfsNodeName = "";
   my $out = `mmgetstate | grep -e arbitrating -e active -e down -e unknown`;
   my @outSplit = split(/\s+/, $out);
   my $gpfsNodeNumber = @outSplit[1];
   my @sdrout = `mmsdrquery sdrq_node_info all`;
   foreach my $line (@sdrout) 
   {
      my @sdrOutSplit = split(/:/, $line);
      if ($gpfsNodeNumber == @sdrOutSplit[6]) {
         $gpfsNodeName = @sdrOutSplit[8];
         last;
      }
   }
 
   my $owner = `dsmmigfs q -d $fsMntPt | grep 'Owner Node:' | awk '{print \$3}'`;
   chomp($owner);
   if( $owner ne $gpfsNodeName ) {
      die "ERROR : $gpfsNodeName is not the owner of $fsMntPt\n";
   }
   
   &setPaths;
}

sub setPaths
{
   if( length($tmpDir) == 0 ) {
      die "ERROR : Temporary folder must be specified.\n";
   }
   $ruleOutput  = $tmpDir."/dsmreconcile.gpfs.$startTime";
   $ruleFile    = $tmpDir."/dsmreconcile.gpfs.$startTime.policy.rule";
   $scanOutFile = $tmpDir."/dsmreconcile.gpfs.$startTime.list.$scanOut";
   $prepOutput  = $tmpDir."/dsmreconcile.gpfs.$startTime.list.prep.out";
   $reorgOutput = $tmpDir."/dsmreconcile.gpfs.$startTime.list.reorg.out";
}

sub verifyMntPoint
{
   my $found = 0;
   print "\nverify mounting point\n" if ($debug > 0);
   
   my $platform = `uname`;
   chomp $platform;
   print "platform is $platform\n" if ($debug > 0);

   open(MNTOUT, "mount |")
   or die "mount output query failed: $!\n";
   
   while(<MNTOUT>)
   {
      if ( ( ( $platform eq "Linux" ) && ( "$_" =~ "^(.+) on ([^\\s]+) type gpfs " ) ) ||
           ( ( $platform eq "AIX" ) && ( "$_" =~ /^\s+(\/dev\/\S+)\s+(\/\S+)\s+mmfs\s+.*/ ) ) ) {
         print "$_" if ($debug > 1);
         if( "$2" eq "$fsMntPt" ) {
            $found = 1;
            last;
         }
      }
   }
   close(MNTOUT);

   return $found;
}

sub trDevToMntPoint
{
   print "\ndev to mounting point\n" if ($debug > 0);
   
   open(MNTOUT, "mount |")
   or die "mount output query failed: $!\n";
   
   while(<MNTOUT>)
   {
      if( "$_" =~ "^(\/dev\/.+) on ([^\\s]+) type gpfs ") {
         print "$_" if ($debug > 1);
         if( "$1" eq "$fsName" or "$1" eq '/dev/'.$fsName ) {
            $fsMntPt = $2;
            last;
         }
      }
   }
   close(MNTOUT);

   if( length($fsMntPt) == 0 ) {
      die "ERROR : Could not identify mounting point for $fsName\n";
   }
}

sub createRule
{
   open(RULE, ">$ruleFile")
   or die "Can't open $ruleFile: $!\n";
   
   print RULE "RULE EXTERNAL LIST '$scanOut' EXEC ''\n";
   print RULE "RULE 'Reconcile' LIST '$scanOut'\n";
   print RULE "    SHOW( CASE WHEN XATTR('dmapi.IBMObj') IS NULL THEN ('P ') ELSE ('M ') END || XATTR('dmapi.IBMexID') )\n";
   print RULE "    WHERE XATTR('dmapi.IBMexID') IS NOT NULL\n";
   
   close(RULE)
   or die "Can't close $ruleFile: $!\n";
}

sub createServerSpecRule
{
   open(RULE, ">$ruleFile")
   or die "Can't open $ruleFile: $!\n";
   
   print RULE "define(server,$serverName)\n\n";
   print RULE "RULE EXTERNAL LIST '$scanOut' EXEC ''\n";
   print RULE "RULE 'Reconcile' LIST '$scanOut'\n";
   print RULE "    SHOW( CASE WHEN XATTR('dmapi.IBMObj') IS NULL THEN ('P ') ELSE ('M ') END || XATTR('dmapi.IBMexID') )\n";
   print RULE "    WHERE XATTR('dmapi.IBMexID') IS NOT NULL\n";
   print RULE "    AND   XATTR('dmapi.IBMServ') LIKE 'server%'\n";
   
   close(RULE)
   or die "Can't close $ruleFile: $!\n";
}

sub invokePolicyEngine
{
   my $policyCommand;

   $policyCommand = "/usr/lpp/mmfs/bin/mmapplypolicy";
   $policyCommand = $policyCommand." $fsName -P $ruleFile -s $tmpDir";
   $policyCommand = $policyCommand." -f $ruleOutput -I defer";
   
   print "\n$policyCommand\n" if ($debug > 0);
   
   open(POLICY, "$policyCommand |")
   or die "Can't open $policyCommand: $!\n";
   
   while(<POLICY>)
   {
      print "$_" if ($debug > 1);
   }
   close(POLICY);

   # in the case that there are no pre/migrated file existing we need to ensure
   # the scan output file exists
   open(OUT, ">>$scanOutFile")
   or die "Can't open $scanOutFile: $!";
   close(OUT);
}

sub prepFilelist
{
   my $prepCommand;

   # given format from GPFS            : INODE IGEN SNAPID MIGSTATE IBMEXOBJID -- PATH
   # expected format from dsmreconcile : IBMEXOBJID INODE IGEN MIGSTATE PATH

   if( ! -f $scanOutFile ) {
      print "No scan output created, nothing to do.\n";
      return(2);
   }
   
   # reorganize the line structure
   open(IN, "< $scanOutFile")
   or die "Can't open $scanOutFile: $!";
   
   open(OUT, ">$reorgOutput")
   or die "Can't open $reorgOutput: $!";
   
   while (<IN>) 
   {
      print "Input: $_" if ($debug > 2);
      if (/^(\d+) (\d+) \d+\s+(M|P)\s+(\w+) -- (.*)$/s)
      {
         print OUT "$4 $1 $2 $3 $5";
         print "Output: $4 $1 $2 $3 $5" if ($debug > 2);
         $nFiles += 1;
      }
      else
      {
         print "Invalid filelist record: $_\n";
      }
   }
   close(IN);
   close(OUT);
   
   # sort the file list and change the \n to \0.
   $prepCommand = "sort -T \"$tmpDir\" $reorgOutput |";
   $prepCommand = $prepCommand." tr -s '\\n' '\\0' > $prepOutput";
   
   print "\n$prepCommand\n" if ($debug > 0);
   
   open(PREP, "$prepCommand |")
   or die "Can't open $prepCommand: $!\n";
   
   while(<PREP>)
   {
      print "$_" if ($debug > 1);
   }
   close(PREP);
   return(0);
}

sub invokeReconcileServerSpec
{
   my $dsmreconcileCommand = "dsmreconcile -o -d -server=$serverName -filelist=$prepOutput $fsMntPt 2>&1";
   
   print "\n$dsmreconcileCommand\n" if ($debug > 0);
   
   open(RECONCILE, "$dsmreconcileCommand |")
   or die "Can't open $dsmreconcileCommand: $!\n";
   
   print "\n";
   while(<RECONCILE>)
   {
      if ( $_ =~ m/^(ANS.{4}W).*/ )
      {
         $returnCode = 8;
      }
      die "$_" if $_ =~ m/^(ANS.{4}E).*/;
      print "$_";
   }
   close(RECONCILE);
}

sub invokeReconcile
{
   my $dsmreconcileCommand = "dsmreconcile -o -d -filelist=$prepOutput $fsMntPt 2>&1";
   
   print "\n$dsmreconcileCommand\n" if ($debug > 0);
   
   open(RECONCILE, "$dsmreconcileCommand |")
   or die "Can't open $dsmreconcileCommand: $!\n";
   
   print "\n";
   while(<RECONCILE>)
   {
      if ( $_ =~ m/^(ANS.{4}W).*/ )
      {
         $returnCode = 8;
      }
      die "$_" if $_ =~ m/^(ANS.{4}E).*/;
      print "$_";
   }
   close(RECONCILE);
}

# run main programm ###########################################################

sub main
{
   my $rc;

   &verifyArguments;
   
   print "\n## dsmreconcile starts at  : ".(localtime)." ####\n\n";
   print "## gpfs rule file          : $ruleFile\n";
   print "## scan output             : $scanOutFile\n";
   print "## reorganized output      : $reorgOutput\n";
   print "## prep output             : $prepOutput\n";
   
   print "\n## create policy rule      : ".(localtime)." ####\n";
   
   if($useServerSpecRule == 0)
   {
      &createRule;
   }
   else
   {
      &createServerSpecRule;
   }
   
   print "\n## invoke policy engine    : ".(localtime)." ####\n";
   
   &invokePolicyEngine;
   
   print "\n## prepare / sort filelist : ".(localtime)." ####\n";
   
   $rc = prepFilelist();
   print "\n## number of files         : ".$nFiles."\n";
   if( $rc == 0 ) {
      print "\n## invoke dsmreconcile     : ".(localtime)." ####\n";
      
      if($useServerSpecRule == 0) {
         &invokeReconcile;
      } else {
         &invokeReconcileServerSpec;
      }    
      
      print "\n## dsmreconcile ends at    : ".(localtime)." ####\n\n";
   }

   if($debug == 0)
   {
      unlink $ruleFile     if (-f $ruleFile);
      unlink $scanOutFile  if (-f $scanOutFile);
      unlink $reorgOutput  if (-f $reorgOutput);
      unlink $prepOutput   if (-f $prepOutput);
   }
}

eval { &main; };
if($@)
{
   print "$@";
   exit (12);
}
exit ($returnCode);






